#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Solution to Exercises 1,2,3 - Python
# Course "Sicurezza", undergraduate degree in Computer Science
# Ca' Foscari University, Venice
# https://secgroup.dais.unive.it/teaching/sicurezza/
import string

########### Exercise 1 #############

# Reverse using split, join and slicing with step -1
def reverse(t):
	return ' '.join(t.split()[::-1])

caesar_ciphertext ='''
Lq fubswrjudskb, d Fdhvdu flskhu, dovr nqrzq dv Fdhvdu'v flskhu, wkh vkliw
flskhu, Fdhvdu'v frgh ru Fdhvdu vkliw, lv rqh ri wkh vlpsohvw dqg prvw zlghob
nqrzq hqfubswlrq whfkqltxhv. Lw lv d wbsh ri vxevwlwxwlrq flskhu lq zklfk hdfk
ohwwhu lq wkh sodlqwhaw lv uhsodfhg eb d ohwwhu vrph ilahg qxpehu ri srvlwlrqv
grzq wkh doskdehw. Iru hadpsoh, zlwk d ohiw vkliw ri 3, G zrxog eh uhsodfhg eb
D, H zrxog ehfrph E, dqg vr rq. Wkh phwkrg lv qdphg diwhu Mxolxv Fdhvdu, zkr
xvhg lw lq klv sulydwh fruuhvsrqghqfh'''

########### Exercise 2 #############

def caesar(t):
	'''Base solution with all explicit cases.

	This base solution brings letters to the interval 0-25 by subtracting
	ord('A') for uppercase and ord('a') for lowercase. Then it performs
	decryption by subtracting -3 modulo 26. Finally it adds again ord('A') 
	or ord('a') to go back to the right ascii code.

	Assumes letters are ascii (i.e. 0 <= ord < 128).

	Returns:
		plaintext
	'''
	ret = ''
	for c in t:
		if c in string.ascii_uppercase:
		    ret+= chr((ord(c)-ord('A')-3)%26+ord('A'))
		elif c in string.ascii_lowercase:
		    ret+= chr((ord(c)-ord('a')-3)%26+ord('a'))
		else:   
		    ret+= c
	return ret

def caesar_bis(t):
	'''As previous but unifies cases using a variable for ord('A') and ord('a').
	'''
	ret = ''
	for c in t:
		if c not in string.ascii_letters:
			ret += c
		else:
			# computes what should be subtracted
			s = ord('A') if c in string.ascii_uppercase else ord('a')
			# single expression for uppercase and lowercase
			ret += chr(s + (ord(c)-s-3)%26)
		
	return ret

def caesar2(t):
	'''Functional version, all cases in the comprehension.
	'''

	return ''.join(
		[	chr((ord(c)-ord('A')-3)%26+ord('A')) if c in string.ascii_uppercase else
			chr((ord(c)-ord('a')-3)%26+ord('a')) if c in string.ascii_lowercase else
			c
			for c in t
		])

def caesar2_bis(t):
	''' Functional version with an extra function for ord('A') and ord('a').
	'''
	def base(c):
		return ord('A') if c in string.ascii_uppercase else ord('a')

	return ''.join([chr(base(c) + (ord(c)-base(c)-3)%26) if c in string.ascii_letters
		            else 
		            c 
		            for c in t])


########### Exercise 3 #############

# Frequency analysis with exception handling
def frequency(t):
	f = {}
	for c in t:
		try:
			f[c] += 1    # add one occurrence
		except KeyError: # if key c is not in the dictionary
			f[c] = 1	 # add it with frequency 1
	# the lambda selects the second member of the pairs
	# produced by f.items(), i.e., the value
	return sorted(f.items(), key=lambda x: x[1], reverse=True)

# Frequency analysis with check for membership
def frequency2(t):
	f = {}
	for c in t:
		if c in f:		# if key c is in the dictionary
			f[c] += 1	# add one occurrence
		else: 			# if key c is not in the dictionary
			f[c] = 1	# add it with frequency 1
	# the lambda selects the second member of the pairs
	# produced by f.items(), i.e., the value
	return sorted(f.items(), key=lambda x: x[1], reverse=True)


########### main: run the tests ############
def main():
	print('== Exercise 1 ==')
	print(reverse('This is a test in which we revert the order of word'))

	print('\n== Exercise 2 ==')
	print(caesar2_bis(caesar_ciphertext))

	print('\n== Exercise 3 ==')
	print(frequency2('This is a test in which we count the frequency of letters. Guess what? Blank space is the most frequent!'))

if __name__ == '__main__':
	main()